import glob
import json
from typing import Generator

from hrfh.models import HTTPResponse
from hrfh.utils.parser import create_http_response_from_bytes, create_http_response_from_json


def yield_http_response_from_json(folder, limit=4) -> Generator[HTTPResponse, None, None]:
    for index, path in enumerate(glob.glob(f"{folder}/**/*.json", recursive=True)):
        with open(path) as f:
            if index > limit:
                break
            yield create_http_response_from_json(json.load(f))

def yield_http_response_from_plain(folder, limit=4) -> Generator[HTTPResponse, None, None]:
    for index, path in enumerate(glob.glob(f"{folder}/**/*.txt", recursive=True)):
        with open(path, mode="rb") as f:
            if index > limit:
                break
            yield create_http_response_from_bytes(f.read())



from typing import List, Tuple

from bs4 import BeautifulSoup

from hrfh.utils.hash import sha256sum
from hrfh.utils.mask import mask_sentence
from hrfh.utils.tokenize import tokenize_html

from dataclasses import dataclass, field
from typing import List, Tuple

@dataclass
class HTTPResponse:
    ip: str = ""
    port: int = 80
    version: str = "HTTP/1.1"
    status_code: int = 200
    status_reason: str = "OK"
    headers: List[Tuple[str, str]] = field(default_factory=list)
    body: bytes = b""

    def __post_init__(self):
        if self.version == 10:
            self.version = "HTTP/1.0"
        if self.version == 11:
            self.version = "HTTP/1.1"
        self.masked: str = self._mask()

    def __repr__(self) -> str:
        return f"<HTTPResponse {self.ip}:{self.port} {self.status_code} {self.status_reason}>"

    def _mask(self) -> str:
        return self._preprocess()

    def dump(self) -> str:
        lines = [
            f"{self.version} {self.status_code} {self.status_reason}"
        ]
        for key, value in self.headers:
            lines.append(f"{key}: {value}")
        lines.append("")
        lines.append(self.body.decode("utf-8"))
        return "\r\n".join(lines)

    def fuzzy_hash(self, hasher=sha256sum) -> str:
        return hasher(self.masked)
    
    def get_tokenized_body(self) -> List[str]:
        soup = BeautifulSoup(self.body, "html.parser")
        masked_html_tokens = []
        for token in tokenize_html(soup):
            if token.startswith("<") and token.endswith(">"):
                # append html tags
                masked_html_tokens.append(token)
            else:
                # append masked text content
                # TODO: handle random string in javascript by create a abstract syntax tree [1] for <script> tag
                # [1] https://github.com/tree-sitter/py-tree-sitter
                masked_html_tokens.append(mask_sentence(token))
        return masked_html_tokens

    def _preprocess(self) -> str:
        header_lines = []
        strip_headers = [
            "Expires",
            "Date",
            "Content-Length",
            "Location",
            "Via",
            "via",
            "Last-Modified",
        ]
        shoud_not_mask_headers = [
            "Server",
            "Connection",
            "Content-Type",
            "Content-Encoding",
            "Cache-Control",
        ]
        headers = []
        if type(self.headers) is dict:
            headers = self.headers.items()
        else:
            headers = self.headers
        for key, value in headers:
            if key in strip_headers:
                value = "REMOVED"
            elif key in shoud_not_mask_headers:
                value = value
            else:
                value = mask_sentence(value)
            header_lines.append(f"{key}: {value}")
        lines = [f"{self.version} {self.status_code} {self.status_reason}"]
        lines += sorted(header_lines)
        lines.append("")
        lines += self.get_tokenized_body()
        return "\r\n".join(lines)



class TrieNode:
    def __init__(self):
        self.children = {}
        self.is_end_of_word = False

class Trie:
    def __init__(self):
        self.root = TrieNode()

    def insert(self, word):
        node = self.root
        for char in word:
            if char not in node.children:
                node.children[char] = TrieNode()
            node = node.children[char]
        node.is_end_of_word = True

    def search(self, word):
        node = self.root
        for char in word:
            if char not in node.children:
                return False
            node = node.children[char]
        return node.is_end_of_word

    def starts_with(self, prefix):
        node = self.root
        for char in prefix:
            if char not in node.children:
                return False
            node = node.children[char]
        return True



import math
from collections import Counter

from hrfh.utils.mask import is_common_word


def shannon_char_entropy(data: str):
    return shannon_ngram_entropy(data, n=1)


def normalized_shannon_entropy(data: str):
    return shannon_char_entropy(data) / math.log(len(data), 2)


def shannon_ngram_entropy(word: str, n: int):
    ngrams = [word[i : i + n] for i in range(len(word) - n + 1)]
    freq = Counter(ngrams)
    total_ngrams = len(ngrams)
    return -sum(
        (count / total_ngrams) * math.log2(count / total_ngrams)
        for count in freq.values()
    )


def is_gibberish(token: str, n: int = 2, entropy_threshold: int = 3):
    """
    >>> is_gibberish("apple")
    False
    >>> is_gibberish("apples")
    False
    >>> is_gibberish("applicable")
    False
    >>> is_gibberish("f638055d-afea-4249-bea7-c73b2a4fec5f")
    True
    >>> is_gibberish("ecea031b74f3e08cef1fba8765196145")
    True
    """
    if is_common_word(token):
        return False
    entropy = shannon_ngram_entropy(token, n)
    return entropy > entropy_threshold



import http.client
import io

from hrfh.models import HTTPResponse


class FakeSocket:
    def __init__(self, bytes_stream):
        self._file = bytes_stream

    def makefile(self, *args, **kwargs):
        return self._file


def create_http_response_from_bytes(data: bytes) -> HTTPResponse:
    response_stream = FakeSocket(io.BytesIO(data))
    response = http.client.HTTPResponse(response_stream)
    response.begin()
    return HTTPResponse(
        ip="1.1.1.1",
        port=80,
        version=response.version,
        status_code=response.status,
        status_reason=response.reason,
        # NOTE: the order of headers will be lost if we insist to use response.getheaders()
        headers=response.getheaders(),
        body=response.read(),
    )


def create_http_response_from_json(data: dict):
    return HTTPResponse(
        ip=data.get("ip"),
        port=data.get("port", 80),
        status_code=data.get("status_code"),
        status_reason=data.get("status_reason"),
        headers=data.get("headers"),
        body=data.get("body"),
    )



def tokenize_html(node):
    tokens = []

    def _tokenize(node):
        if node.name and node.name != '[document]':
            tokens.append(f"<{node.name}>")
            for child in node.children:
                _tokenize(child)
            tokens.append(f"</{node.name}>")
        elif node.string and node.string.strip():
            tokens.append(node.string.strip())

    for child in node.children:
        _tokenize(child)
    
    return tokens


import re

from nltk import download
from nltk.corpus import words
from nltk.stem import WordNetLemmatizer
from nltk.tokenize import word_tokenize

download("wordnet")
download("words")
download("punkt")

lemmatizer = WordNetLemmatizer()
common_words = set(w.lower() for w in words.words())


def is_common_word(token: str) -> bool:
    return lemmatizer.lemmatize(token.lower()) in common_words


def is_hex_digit(c: str) -> bool:
    return c in "0123456789abcdefABCDEF"


def is_hex_string(token: str) -> bool:
    return all(is_hex_digit(c) for c in token)


def is_digit_string(token: str) -> bool:
    return all(c.isdigit() for c in token)


def is_base64_string(token: str) -> bool:
    return all(c.isalnum() or c in "+/=" for c in token)


def is_uncommon_word(token: str) -> bool:
    return not is_common_word(token)


def is_random_string(token: str) -> bool:
    if (
        (re.search(r"\d", token) and len(token) > 2)
        or re.search(r"\W", token)
        or len(token) > 15
    ):
        return True
    return False


def need_mask(token: str) -> bool:
    if is_common_word(token):
        return False
    if is_hex_string(token):
        return True
    if is_digit_string(token):
        return True
    if is_base64_string(token):
        return True
    if is_random_string(token):
        return True
    return False


def mask_word(word: str) -> str:
    mask = "MASKED"
    return mask if is_random_string(word) else word


def mask_sentence(sentence: str) -> str:
    return " ".join([mask_word(word) for word in word_tokenize(sentence)])
